/**** Yet Another Alpha Simulator: SimAlpha        by Kenji KISE ****/
/********************************************************************/
#include "define.h"

/* 64 bit multiply, Input of A and B, output of High 64b and Low 64b. */
/**********************************************************************/
void mul128(uint64_t a, uint64_t b, uint64_t *rh, uint64_t *rl){
  uint64_t ah,al, bh, bl;
  uint64_t x,y,z;
  int carry=0;  /* has a value of 0, 1 and 2 */
  
  ah = a >> 32;
  al = a & 0xFFFFFFFF;
  
  bh = b >> 32;
  bl = b & 0xFFFFFFFF;

  x = al*bl;
  y = (al*bh << 32);
  z = x + y;
  if(z<x || z<y) carry ++;
  x = z;
  y = (ah*bl << 32);
  z = x + y;
  if(z<x || z<y) carry ++;
  *rl = z;
  
  *rh = ah*bh+ (al*bh >> 32) + (ah*bl >> 32) + carry;
}

/* For the branch instruction, calculate branch taken or not.         */
/**********************************************************************/
int Condition(int code, DATA_TYPE Rav){
  int taken=0;
  double Fav=0.0;
  //  signal(SIGFPE, (void *)signal_float_branch);
  switch(code){
  case 0x0: taken = 1;                 break;
  case 0x4: taken = 1;                 break;
  case 0x9: taken = (Rav==0);          break;
  case 0xe: taken = ((int64_t)Rav>=0); break;
  case 0xf: taken = ((int64_t)Rav>0);  break;
  case 0x8: taken = (!(Rav & 1));      break;
  case 0xc: taken = (  Rav & 1 );      break;
  case 0xb: taken = ((int64_t)Rav<=0); break;
  case 0xa: taken = ((int64_t)Rav<0 ); break;
  case 0xd: taken = (Rav!=0);          break;
    
  case 0x1: memcpy(&Fav,&Rav,8); taken = (Fav==0.0); break;
  case 0x2: memcpy(&Fav,&Rav,8); taken = (Fav< 0.0); break;
  case 0x3: memcpy(&Fav,&Rav,8); taken = (Fav<=0.0); break;
  case 0x5: memcpy(&Fav,&Rav,8); taken = (Fav!=0.0); break;
  case 0x6: memcpy(&Fav,&Rav,8); taken = (Fav>=0.0); break;
  case 0x7: memcpy(&Fav,&Rav,8); taken = (Fav> 0.0); break;
  default: fprintf(stderr, "*** ex_branch default error!\n");
  }
  //  printf("code=4'h%x; Rav=64'h%016qx; result=%d; #10;\n", 
  //	 code, Rav, taken);
  //  signal(SIGFPE, SIG_IGN);
  return taken;
}

/* Jump & Branch Unit, Update TPC                                     */
/**********************************************************************/
int BRU(INST_TYPE ir, data_t *R1, data_t *R2, data_t *npc, data_t *tpc){
  int Op = (ir>>26) & 0x3F;
  tpc->st(Condition(Op&0xF, R1->ld()) ? 
	  npc->ld() + R2->ld() :  npc->ld());
  
  return 0;
}

/**********************************************************************/
int MUL_value(INST_TYPE ir, 
	      DATA_TYPE Rav, DATA_TYPE Rbv, DATA_TYPE *W){
  DATA_TYPE Rcv, tmp2, tmp3;
  int func7 = (ir>>5) & 0x7f;

  switch(func7)
    {
    case 0x00: /*__MULL; */
      Rcv = (Rav*Rbv) & 0x00000000FFFFFFFFllu;
      if(Rcv & BIT31) Rcv |= EXTND32;
      break;
      
    case 0x20: /*__MULQ; */
      mul128(Rav, Rbv, &tmp2, &tmp3);
      Rcv = tmp3;
      break;
      
    case 0x30: /*__UMULH; */
      mul128(Rav, Rbv, &tmp2, &tmp3);
      Rcv = tmp2;
      break;
      
    default:
      printf("** Warning Code 0x13 default. func11 %x\n",func7);
      Rcv = 0;
    }

  *W = Rcv;
  return 0;
}

/* Rc <- BYTE_ZAP(Rav, byte_mask)                                     */
/**********************************************************************/
uint64_t byte_zap(uint64_t Rav, int byte_mask){
  if(byte_mask & BIT0) Rav &=  ZAP0;
  if(byte_mask & BIT1) Rav &=  ZAP1;
  if(byte_mask & BIT2) Rav &=  ZAP2;
  if(byte_mask & BIT3) Rav &=  ZAP3;
  if(byte_mask & BIT4) Rav &=  ZAP4;
  if(byte_mask & BIT5) Rav &=  ZAP5;
  if(byte_mask & BIT6) Rav &=  ZAP6;
  if(byte_mask & BIT7) Rav &=  ZAP7;
  return Rav;
}


/* Rc <- BYTE_ZAP(Rav, byte_mask)                                     */
/**********************************************************************/
int ShiftUnit_value(INST_TYPE ir, 
		    DATA_TYPE Rav, DATA_TYPE Rbv, DATA_TYPE *R){
  int func7 = (ir>>5) & 0x7f;
  DATA_TYPE Rcv=0;
  unsigned int byte_mask, byte_loc;
  
  switch(func7){
  case 0x34: Rcv=(uint64_t)Rav >> (Rbv & 0x3F); break; // SRL
  case 0x39: Rcv=(uint64_t)Rav << (Rbv & 0x3F); break; // SLL
  case 0x3c: Rcv=( int64_t)Rav >> (Rbv & 0x3F); break; // SRA
    
  case 0x06: /*__EXTBL; */
    Rcv = (uint64_t)Rav >> ((Rbv & 0x7)<<3);
    Rcv &= 0x00000000000000FFllu;
    break;
  case 0x16: /*__EXTWL; */
    Rcv = (uint64_t)Rav >> ((Rbv & 0x7)<<3);
    Rcv &= 0x000000000000FFFFllu;
    break;
  case 0x26: /*__EXTLL; */
    Rcv = (uint64_t)Rav >> ((Rbv & 0x7)<<3);
    Rcv &= 0x00000000FFFFFFFFllu;
    break;
    
  case 0x36: /*__EXTQL; */
    Rcv = (uint64_t)Rav >> ((Rbv & 0x7)<<3);
    break;
  case 0x5a: /*__EXTWH; */
    Rcv = (uint64_t)Rav << (64-((Rbv & 0x7)<<3));
    Rcv &= 0x000000000000FFFFllu;
    break;
  case 0x6a: /*__EXTLH; */
    Rcv = (uint64_t)Rav << (64-((Rbv & 0x7)<<3));
    Rcv &= 0x00000000FFFFFFFFllu;
    break;
  case 0x7a: /*__EXTQH; */
    Rcv = (uint64_t)Rav << (64-((Rbv & 0x7)<<3));
    break;
    
  case 0x0b: /*__INSBL; */
    byte_mask = 0x00000001ull << (Rbv & 0x7);
    byte_loc  = ((Rbv & 0x7) << 3);
    Rcv = byte_zap((Rav << byte_loc), ~byte_mask);
    break;
  case 0x1b: /*__INSWL; */
    byte_mask = 0x00000003ull << (Rbv & 0x7);
    byte_loc  = ((Rbv & 0x7) << 3);
    Rcv = byte_zap((Rav << byte_loc), ~byte_mask);
    break;
  case 0x2b: /*__INSLL; */
    byte_mask = 0x0000000Full << (Rbv & 0x7);
    byte_loc  = ((Rbv & 0x7) << 3);
    Rcv = byte_zap((Rav << byte_loc), ~byte_mask);
    break;
  case 0x3b: /*__INSQL; */
    byte_mask = 0x000000FFull << (Rbv & 0x7);
    byte_loc  = ((Rbv & 0x7) << 3);
    Rcv = byte_zap((Rav << byte_loc), ~byte_mask);
    break;
    
  case 0x57: /*__INSWH; */
    byte_mask = ((0x00000003ull << (Rbv & 0x7))>>8);
    byte_loc  = (64 - ((Rbv & 0x7)<<3)) & 0x3f;
    Rcv = byte_zap((Rav >> byte_loc), ~byte_mask);
    break;
  case 0x67: /*__INSLH; */
    byte_mask = ((0x0000000Full << (Rbv & 0x7))>>8);
    byte_loc  = (64 - ((Rbv & 0x7)<<3)) & 0x3f;
    Rcv = byte_zap((Rav >> byte_loc), ~byte_mask);
    break;
  case 0x77: /*__INSQH; */
    byte_mask = ((0x000000FFull << (Rbv & 0x7))>>8);
    byte_loc  = (64 - ((Rbv & 0x7)<<3)) & 0x3f;
    Rcv = byte_zap((Rav >> byte_loc), ~byte_mask);
    break;
    
  case 0x02: /*__MSKBL; */
    byte_mask = 0x00000001ull << (Rbv & 0x7);
    Rcv = byte_zap(Rav, byte_mask);
    break;
  case 0x12: /*__MSKWL; */
    byte_mask = 0x00000003ull << (Rbv & 0x7);
    Rcv = byte_zap(Rav, byte_mask);
    break;
  case 0x22: /*__MSKLL; */
    byte_mask = 0x0000000Full << (Rbv & 0x7);
    Rcv = byte_zap(Rav, byte_mask);
    break;
  case 0x32: /*__MSKQL; */
    byte_mask = 0x000000FFull << (Rbv & 0x7);
    Rcv = byte_zap(Rav, byte_mask);
    break;
  case 0x52: /*__MSKWH; */
    byte_mask = (0x00000003ull << (Rbv & 0x7)) >> 8;
    Rcv = byte_zap(Rav, byte_mask);
    break;
  case 0x62: /*__MSKLH; */
    byte_mask = (0x0000000Full << (Rbv & 0x7)) >> 8;
    Rcv = byte_zap(Rav, byte_mask);
    break;
  case 0x72: /*__MSKQH; */
    byte_mask = (0x000000FFull << (Rbv & 0x7)) >> 8;
    Rcv = byte_zap(Rav, byte_mask);
    break;
    
  case 0x30: /*__ZAP; */
    Rcv = byte_zap(Rav,  Rbv);
    break;
  case 0x31: /*__ZAPNOT; */
    Rcv = byte_zap(Rav, ~Rbv);
    break;
    
  default:
    printf("In logic.cc Warning Code 0x12 defalult. func11 0x%u\n",
	   func7);
    Rcv = 0;
  }

  *R = Rcv;
  return 0;
}




/* Integer Arithmetic and Logic Unit                                  */
/**********************************************************************/
int ALU_value(INST_TYPE ir, DATA_TYPE Rav, int Racv, 
	      DATA_TYPE Rbv, DATA_TYPE *w, int *cmov){
  DATA_TYPE Rcv=0;
  int cm = 0;

  int Op = (ir>>26) & 0x3F;
  int func7 = (ir>>5) & 0x7F;

  switch(Op){
  case 0x0a: Rcv = Rav + Rbv; break; /**** LD ****/
  case 0x0b: Rcv = (Rav + Rbv) & ~7; break;
  case 0x0c: Rcv = Rav + Rbv; break;
  case 0x20: Rcv = Rav + Rbv; break;
  case 0x21: Rcv = Rav + Rbv; break;
  case 0x22: Rcv = Rav + Rbv; break;
  case 0x23: Rcv = Rav + Rbv; break;
  case 0x28: Rcv = Rav + Rbv; break;
  case 0x29: Rcv = Rav + Rbv; break;
  case 0x2a: Rcv = Rav + Rbv; break;
  case 0x2b: Rcv = Rav + Rbv; break;

  case 0x0d: Rcv = Rav + Rbv; break; /**** ST ****/
  case 0x0e: Rcv = Rav + Rbv; break;
  case 0x0f: Rcv = (Rav + Rbv) & ~7; break;
  case 0x24: Rcv = Rav + Rbv; break;
  case 0x25: Rcv = Rav + Rbv; break;
  case 0x26: Rcv = Rav + Rbv; break;
  case 0x27: Rcv = Rav + Rbv; break;
  case 0x2c: Rcv = Rav + Rbv; break;
  case 0x2d: Rcv = Rav + Rbv; break;
  case 0x2e: Rcv = Rav + Rbv; break;
  case 0x2f: Rcv = Rav + Rbv; break;

  case 0x08: Rcv = Rav + Rbv; break; /* LDA  */
  case 0x09: Rcv = Rav + Rbv; break; /* LDAH */

  case OP_INTA:
    switch(func7){
    case 0x02:
    case 0x0b: 
    case 0x22:
    case 0x2b: Rav = (Rav << 2); break;
    case 0x12: 
    case 0x1b: 
    case 0x32:
    case 0x3b: Rav = (Rav << 3); break;
    }

    switch(func7)
      {
      case 0x00: Rcv = Rav + Rbv; break;
      case 0x20: Rcv = Rav + Rbv; break;
      case 0x02: Rcv = Rav + Rbv; break;
      case 0x22: Rcv = Rav + Rbv; break;
      case 0x12: Rcv = Rav + Rbv; break;
      case 0x32: Rcv = Rav + Rbv; break;

      case 0x09: Rcv = Rav - Rbv; break;
      case 0x29: Rcv = Rav - Rbv; break;
      case 0x0b: Rcv = Rav - Rbv; break;
      case 0x2b: Rcv = Rav - Rbv; break;
      case 0x1b: Rcv = Rav - Rbv; break;
      case 0x3b: Rcv = Rav - Rbv; break;

      case 0x2d: Rcv = ((int64_t)Rav==(int64_t)Rbv); break;
      case 0x4d: Rcv = ((int64_t)Rav< (int64_t)Rbv); break;
      case 0x6d: Rcv = ((int64_t)Rav<=(int64_t)Rbv); break;
      case 0x3d: Rcv = (Rav <= Rbv); break;
      case 0x1d: Rcv = (Rav <  Rbv); break;

      case 0x0f: /*__CMPBGE; */
	Rcv=0;
	if((Rav & ~ZAP0) >= (Rbv & ~ZAP0)) Rcv |= BIT0;
	if((Rav & ~ZAP1) >= (Rbv & ~ZAP1)) Rcv |= BIT1;
	if((Rav & ~ZAP2) >= (Rbv & ~ZAP2)) Rcv |= BIT2;
	if((Rav & ~ZAP3) >= (Rbv & ~ZAP3)) Rcv |= BIT3;
	if((Rav & ~ZAP4) >= (Rbv & ~ZAP4)) Rcv |= BIT4;
	if((Rav & ~ZAP5) >= (Rbv & ~ZAP5)) Rcv |= BIT5;
	if((Rav & ~ZAP6) >= (Rbv & ~ZAP6)) Rcv |= BIT6;
	if((Rav & ~ZAP7) >= (Rbv & ~ZAP7)) Rcv |= BIT7;
	break;
	
      default: printf("Warning in 0x10 arithmetic\n");	Rcv=0;
      }

    switch(func7){
    case 0x00: 
    case 0x09:
    case 0x02:
    case 0x12:
    case 0x0b:
    case 0x1b: if(Rcv & BIT31) Rcv |= EXTND32; else Rcv &= MASK32;
      break;
    }
    break;
    
  case OP_INTL:
    switch(func7){
    case 0x08: 
    case 0x28: 
    case 0x48: Rbv = ~Rbv; break;
    }

    switch(func7){
    case 0x00: Rcv = Rav & Rbv; break;
    case 0x08: Rcv = Rav & Rbv; break;
    case 0x20: Rcv = Rav | Rbv; break;
    case 0x28: Rcv = Rav | Rbv; break;
    case 0x40: Rcv = Rav ^ Rbv; break;
    case 0x48: Rcv = Rav ^ Rbv; break;

    case 0x14: Rcv=Rbv; cm=(  Rav & 1);       break;
    case 0x16: Rcv=Rbv; cm=(!(Rav & 1));      break;
    case 0x24: Rcv=Rbv; cm=(Rav==0);          break;
    case 0x26: Rcv=Rbv; cm=(Rav!=0);          break;
    case 0x44: Rcv=Rbv; cm=((int64_t)Rav< 0); break;
    case 0x46: Rcv=Rbv; cm=((int64_t)Rav>=0); break;
    case 0x64: Rcv=Rbv; cm=((int64_t)Rav<=0); break;
    case 0x66: Rcv=Rbv; cm=((int64_t)Rav> 0); break;

    case 0x61: Rcv=Rbv & ~CPU_FEATURE_MASK;    break; /* AMASK   */
    case 0x6c: Rcv=IMPLEMENTATION_VERSION;     break; /* IMPLVER */

      /** CMOVb: This code is assigned by Kenji KISE **/
    case 0x6e: Rcv = (Racv) ? Rbv : Rav;   break;

    default:
      printf("Warning 0x11 Func:0x%x Default Case.\n",func7); 
      Rcv=0; break;
    }
    break;
  }

  *w    = Rcv;
  *cmov = cm;
  return 0;
}

/* Floating Arithmetic and Logic Unit                                 */
/**********************************************************************/
int FALU_value(INST_TYPE ir, DATA_TYPE Rav, int Racv, 
	       DATA_TYPE Rbv, DATA_TYPE *w, int *cmov){

  DATA_TYPE Rcv=0;

  int code   = (ir>>26) & 0x3F;
  int func11 = (ir>>5) & 0x7FF;

  double Fav   = 0.0, Fbv   = 0.0, Fcv   = 0.0;
  
  signal(SIGFPE, SIG_IGN); // Ignore Floagint Exceptions

  memcpy(&Fav, &Rav, 8);
  memcpy(&Fbv, &Rbv, 8);
      
  switch(code){

  case OP_ITFP: /*** Integer to floating-point register move ***/
    switch(func11){
    case 0x024: // ITOFT
      //      if(Rav==0) Rav=0xaae4d; /* ????? */
      memcpy(&Fcv, &Rav, 8);
      break;

    case 0x004: // ITOFS
      {
	DATA_TYPE tmp = 0;
	tmp = ((Rav & BIT31) << 32) | 
	  (map_s(Rav) << 52) |
	  ((Rav & 0x7fffff) << 29);
	memcpy(&Fcv, &tmp, 8);
      }
      break;

    case 0x0ab: // SQRTT
      {
	Fcv = sqrt(Fbv);
      }
      break;

    default: 
      printf("Warning 0x14 Floating Default. func11 %x \n", func11);
      exit(1);
      Fcv=0.0;
    }
    memcpy(&Rcv, &Fcv, 8);
    break;

  case OP_FPTI: /*** Floating-point to integer register move ***/
    switch(func11){
    case 0x000: // SEXTB(Sign Extend Byte): Rc <= SEXT(Rbv[7:0])
      Rcv = (Rbv & BIT7) ? Rbv |= EXTND8 : (Rbv & 0xffull);
      break;
    case 0x001: // SEXTW(Sign Extend Word): Rc <= SEXT(Rbv[15:0])
      Rcv = (Rbv & BIT15) ? Rbv |= EXTND16 : (Rbv & 0xffffull);
      break;
    case 0x070: // FTOIT: Rc <= Fav
      Rcv = Rav;
      break;
    case 0x078: // FTOIS: Rc <= Fav
      Rcv = ((Rav>>32) & 0xc0000000) | ((Rav >> 29) & 0x3fffffff);
      if(Rav & BIT63) Rcv |= EXTND32;
      break;

    default: 
      printf("Warning 0x1c Floating Default. func11 %x \n", func11);
      exit(1);
    }
    break;

  case OP_FLTV: /*** VAX floating-point instruction ***/
    printf("Warning 0x15 Floating Default. func11 %x \n", func11);
    exit(1);
    
  case OP_FLTI: { /*** IEEE floating-point instruction ***/
      switch(func11){

      case 0x080:
      case 0x000:
      case 0x040:
      case 0x0c0:
      case 0x180:
      case 0x100:
      case 0x140:
      case 0x1c0: 
      case 0x580: 
      case 0x500: 
      case 0x540:
      case 0x5c0:
      case 0x780:
      case 0x700:
      case 0x740:
      case 0x7c0: Fcv = Fav + Fbv; break; break; // ADDS
	
      case 0x0a0:
      case 0x020:
      case 0x060:
      case 0x0e0:
      case 0x1a0:
      case 0x120:
      case 0x160:
      case 0x1e0: 
      case 0x5a0:
      case 0x520:
      case 0x560:
      case 0x5e0:
      case 0x7a0:
      case 0x720:
      case 0x760:
      case 0x7e0: Fcv = Fav + Fbv; break; // ADDT

      case 0x0a5:
      case 0x5a5: // CMPTEQ
	Fcv = (Fav==Fbv) ? 2.0 : 0.0; break;

      case 0x0a6:	
      case 0x5a6: // CMPTLT
	Fcv = (Fav< Fbv) ? 2.0 : 0.0; break;

      case 0x0a7:
      case 0x5a7: // CMPTLE
	Fcv = (Fav<=Fbv) ? 2.0 : 0.0; break;

      case 0x0a4:
      case 0x5a4: // CMPTUN
	Fcv = (isnan(Fav) || isnan(Fbv)) ? 2.0 : 0.0; break;

      case 0x0bc:
      case 0x03c:
      case 0x07c:
      case 0x0fc: 
      case 0x7bc: 
      case 0x73c: 
      case 0x77c: 
      case 0x7fc: // CVTQS
	Fcv = (double)(int64_t)Rbv;
	break;
	
      case 0x0be:
      case 0x03e:
      case 0x07e:
      case 0x0fe: 
      case 0x7be:
      case 0x73e:
      case 0x77e:
      case 0x7fe: // CVTQT
	Fcv = (double)(int64_t)Rbv;
	//	printf("*** CVTQT  Fcv %f Rbv %llx \n", Fcv, Rbv);
	break;

      case 0x0ac:
      case 0x02c:
      case 0x06c:
      case 0x0ec:
      case 0x1ac:
      case 0x12c:
      case 0x16c:
      case 0x1ec:
      case 0x5ac:
      case 0x52c:
      case 0x56c:
      case 0x5ec:
      case 0x7ac:
      case 0x72c:
      case 0x76c:
      case 0x7ec: // CVTTS
	{
	  memcpy(&Fcv, &Rbv, 8);
	  float tmp = Fcv;
	  Fcv = tmp;
	}
	break;
	
      case 0x083: 
      case 0x003: 
      case 0x043: 
      case 0x0c3: 
      case 0x183: 
      case 0x103: 
      case 0x143: 
      case 0x1c3: 
      case 0x583: 
      case 0x503: 
      case 0x543: 
      case 0x5c3: 
      case 0x783: 
      case 0x703: 
      case 0x743: 
      case 0x7c3: // DIVS
	if(Fbv==0.0){
	  Fcv = 0.0;
	  break;
	}
	Fcv = Fav / Fbv;
	break;

      case 0x0a3:
      case 0x023:
      case 0x063:
      case 0x0e3:
      case 0x1a3:
      case 0x123:
      case 0x163:
      case 0x1e3:
      case 0x5a3:
      case 0x523:
      case 0x563:
      case 0x5e3:
      case 0x7a3:
      case 0x723:
      case 0x763:
      case 0x7e3: // DIVT
	if(Fbv==0.0){
	  Fcv = 0.0;
	  break;
	}
	Fcv = (double)(Fav / Fbv);
	break;

      case 0x082:
      case 0x002:
      case 0x042:
      case 0x0c2:
      case 0x182:
      case 0x102:
      case 0x142:
      case 0x1c2:
      case 0x582:
      case 0x502:
      case 0x542:
      case 0x5c2:
      case 0x782:
      case 0x702:
      case 0x742:
      case 0x7c2: // MULS
	Fcv = Fav * Fbv;   break;

      case 0x0a2:
      case 0x022:
      case 0x062:
      case 0x0e2:
      case 0x1a2:
      case 0x122:
      case 0x162:
      case 0x1e2:
      case 0x5a2:
      case 0x522:
      case 0x562:
      case 0x5e2:
      case 0x7a2:
      case 0x722:
      case 0x762:
      case 0x7e2: // MULT
	Fcv = Fav * Fbv;   break;

      case 0x081:
      case 0x001:
      case 0x041:
      case 0x0c1:
      case 0x181:
      case 0x101:
      case 0x141:
      case 0x1c1:
      case 0x581:
      case 0x501:
      case 0x541:
      case 0x5c1:
      case 0x781:
      case 0x701:
      case 0x741:
      case 0x7c1: // SUBS
	Fcv = Fav - Fbv; break;

      case 0x0a1:
      case 0x021:
      case 0x061:
      case 0x0e1:
      case 0x1a1:
      case 0x121:
      case 0x161:
      case 0x1e1:
      case 0x5a1:
      case 0x521:
      case 0x561:
      case 0x5e1:
      case 0x7a1:
      case 0x721:
      case 0x761:
      case 0x7e1: // SUBT
	Fcv = Fav - Fbv; break;

      case 0x0af:
      case 0x02f:
      case 0x1af:
      case 0x12f:
      case 0x5af:
      case 0x52f: 
      case 0x7af:
      case 0x72f: 
      case 0x0ef:
      case 0x1ef: 
      case 0x5ef:
      case 0x7ef: 
      case 0x06f:
      case 0x16f: 
      case 0x56f:
      case 0x76f: // CVTTQ
	{
	  int64_t tmp;
	  tmp = (int64_t)Fbv;
	  memcpy(&Fcv, &tmp, 8);
	}
	break;

      default: 
	printf("Warning 0x16 Floating Default. func11 %x \n", func11);
	exit(1);
	Fcv=0.0;
      }
      memcpy(&Rcv, &Fcv, 8);

      break;
    }

    
  case OP_FLTL: /*** Floating-point instruction ***/
    {
      switch(func11){
      case 0x020: /*__CPYS;  Copy Sign */
	{ 
	  uint64_t tmp=0;
	  tmp |= (Rav &  BIT63);
	  tmp |= (Rbv & ~BIT63);
	  memcpy(&Fcv, &tmp, 8);
	}
	/*
	tmp  = Rav;
	tmp2 = Rbv;
	tmp2 = ((tmp2 << 1) >> 1);
	if(tmp & BIT63) tmp2 |= BIT63;
	memcpy(&Fcv, &tmp2, 8);
	*/
	break;

      case 0x021: /*__CPYSN; Copy Sign Negative */
	{
	  uint64_t tmp=0;
	  tmp |= (~Rav &  BIT63);
	  tmp |= ( Rbv & ~BIT63);
	  memcpy(&Fcv, &tmp, 8);
	}
	break;

      case 0x022: /*__CPYSE; Copy Sign and Exponent */
	{
	  uint64_t tmp=0;
	  tmp |= (Rav &  0xFFF0000000000000llu);
	  tmp |= (Rbv & ~0xFFF0000000000000llu);
	  memcpy(&Fcv, &tmp, 8);
	}
	/*
	tmp  = Rav;
	tmp2 = Rbv;
	tmp  &= 0xFFF0000000000000llu;
	tmp2 &= 0x000FFFFFFFFFFFFFllu;
	tmp |= tmp2;
	memcpy(&Fcv, &tmp, 8);
	*/
	break;
	
      case 0x010: /*__CVTLQ; */
	{
	  uint64_t tmp = 0;
	  tmp  = 0;
	  tmp |= ((Rbv>>32) & 0xc0000000ull);
	  tmp |= ((Rbv>>29) & 0x3fffffffull);
	  if(tmp & BIT31) tmp |= EXTND32;
	  memcpy(&Fcv, &tmp, 8);
	}
	break;

      case 0x030: /*__CVTQL; */
	{
	  DATA_TYPE tmp = 0;
	  /* Next line is needed. SimpleScalar has a bug?  */
	  tmp |=  (((Rbv << 32) & 0xc000000000000000ull));
	  tmp |=  (((Rbv << 29) & 0x07ffffffe0000000ull));
	  memcpy(&Fcv, &tmp, 8);
	  break;
	}

      case 0x24: /* __MT_FPCR */
	/* Move Fav to Floating-point Control Register
	   Return 0.0 for the simplicity.
	*/
	Fcv=0.0; break;
	
      case 0x25: /* __MF_FPCR */
	/* Move from Floating-point Control Register to Fa
	   Return 0.0 for the simplicity.
	*/
	Fcv=0.0; break;
	
      case 0x02a:/*__FCMOVEQ; */
	Fcv = Fbv; *cmov = (Fav==0.0);
	break;

      case 0x02b:/*__FCMOVNE; */
	Fcv = Fbv; *cmov=(Fav!=0.0);
	break;

	/** CMOVb: This is assigned by Kenji KISE. **/
      case 0x333:
	Fcv = (Racv) ? Fbv : Fav;
	//	printf("** Condition %d  Fb Fa %qx %qx\n", R1->cmov, Fbv, Fav);
	break;
      
      default:
	printf("Warning Floating 0x17 code default: func11 0x%x\n", func11);
	exit(1);
	Fcv=0.0; break;
      }
      memcpy(&Rcv, &Fcv, 8);
      break;
    }
  }

  *w = Rcv;
  return 0;
}

/* calculate Rcv                                                      */
/**********************************************************************/
int ALU(INST_TYPE ir, data_t *R1, data_t *R2, data_t *Rw){
  DATA_TYPE Rcv =0;
  DATA_TYPE Rav =R1->ld();
  DATA_TYPE Rbv =R2->ld();
  int Racm = R1->cmov;
  int cm=0;

  int Op = (ir>>26) & 0x3F;

  switch(Op){
  case OP_JSR:

  case 0x30: /** BR **/
  case 0x31:
  case 0x32:
  case 0x33:
  case 0x34:
  case 0x35:
  case 0x36:
  case 0x37:
  case 0x38:
  case 0x39:
  case 0x3a:
  case 0x3b:
  case 0x3c:
  case 0x3d:
  case 0x3e:
  case 0x3f:    Rcv = Rbv; break;

  case OP_PAL : Rcv=0; break;
  case OP_MISC: Rcv=0; break;
  case OP_INTS: ShiftUnit_value(ir, Rav, Rbv, &Rcv); break;
  case OP_INTM: MUL_value(ir, Rav, Rbv, &Rcv); break;
  case OP_LDA :
  case OP_LDAH:
  case OP_INTA:
  case OP_INTL: ALU_value(ir, Rav, Racm, Rbv, &Rcv, &cm); break;
  case OP_ITFP:
  case OP_FLTV:
  case OP_FLTI:
  case OP_FLTL:
  case OP_FPTI: FALU_value(ir, Rav, Racm, Rbv, &Rcv, &cm); break;
  default:      ALU_value(ir, Rav, Racm, Rbv, &Rcv, &cm); break;
  }

  Rw->cmov = cm;
  Rw->st(Rcv);
  
  return 0;
}
